/*
 * Copyright 2019 by Kappich Systemberatung Aachen
 * ALL RIGHTS RESERVED.
 */

package de.bsvrz.dav.daf.accessControl.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Hilfsklasse, die von einem DataLoader-Objekt die referenzierten Unterklassen ermittelt und auf eventuelle Rekursionsprobleme hinweist
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
class ChildrenTreeEnumerator {

	private final List<DataLoader> _alreadyVisitedList = new ArrayList<>();

	private final List<DataLoader> _alreadyFinishedList = new ArrayList<>();

	private final DafAccessControlManager _accessControlManager;

	private final DataLoader _node;

	/**
	 * @param accessControlManager AccessControlManager, der über eventuelle Rekursionsprobleme informiert wird
	 * @param node                 Objekt das nach Kindelementen gefragt wird
	 */
	public ChildrenTreeEnumerator(final DafAccessControlManager accessControlManager, final DataLoader node) {
		_accessControlManager = accessControlManager;
		_node = node;
	}

	/**
	 * Gibt alle Kindelemente eines Objekts zurück
	 *
	 * @return Liste mit Kindelementen
	 */
	protected List<DataLoader> enumerateChildren() {
		enumerateChildrenInternal(_node);
		return Collections.unmodifiableList(_alreadyFinishedList);
	}

	/**
	 * Interne rekursiv aufgerufene Hilfsfunktion zum Auflisten von Kindelementen
	 *
	 * @param node Objekt das nach Kindelementen gefragt wird
	 */
	private void enumerateChildrenInternal(DataLoader node) {
		if(_alreadyFinishedList.contains(node)) return;
		if(_alreadyVisitedList.contains(node)) {
			final ArrayList<DataLoader> trace = new ArrayList<>(_alreadyVisitedList);
			trace.removeAll(_alreadyFinishedList);
			trace.add(node);
			_accessControlManager.notifyInfiniteRecursion(node, _alreadyVisitedList.get(_alreadyVisitedList.size() - 1), trace);
		}
		else {
			_alreadyVisitedList.add(node);
			for(DataLoader dataLoader : node.getChildObjects()) {
				enumerateChildrenInternal(dataLoader);
			}
		}
		_alreadyFinishedList.add(node);
	}

	@Override
	public String toString() {
		return "ChildrenTreeEnumerator{" + "_node=" + _node + '}';
	}
}
